@@ -10,11 +10,19 @@
 
 #define RS_XML(a)  RS_XML_##a
 
+#ifdef FROM_GNOME_XML_DIR
+#include <gnome-xml/valid.h>
+#include <gnome-xml/parser.h> 
+#include <gnome-xml/parserInternals.h> 
+#include <gnome-xml/tree.h>
+#include <gnome-xml/entities.h>
+#else
 #include <libxml/valid.h>
 #include <libxml/parser.h> 
 #include <libxml/parserInternals.h> 
 #include <libxml/tree.h>
 #include <libxml/entities.h>
+#endif
 
 USER_OBJECT_ RS_XML(getDTD)(USER_OBJECT_ dtdFileName, USER_OBJECT_ externalId,  USER_OBJECT_ asText, USER_OBJECT_ isURL);
 USER_OBJECT_ RS_XML(createDTDElement)(xmlElementPtr el);
